from datetime import datetime
import boto3
from botocore.exceptions import ClientError
import pandas as pd
from io import BytesIO
from pyarrow.feather import write_feather

from jardin import config as config
from jardin.cache_stores.base import Base
class S3(Base):

    EXTENSION = '.feather'
    
    def __init__(self, bucket_name, path="", delete_expired_files=False):
        self.bucket_name = bucket_name
        self.path = f"{path}/jardin_cache"
        self.delete_expired_files = delete_expired_files
        self.s3 = boto3.resource('s3')
        self.check_valid_bucket()
        

    def check_valid_bucket(self):
        s3_client = boto3.client('s3')
        try:
            s3_client.head_bucket(Bucket=self.bucket_name)
        except ClientError as ex:
            config.logging.warning(f"Bucket {self.bucket_name} does not exist")
            raise ex
            
    def __getitem__(self, key):
        if key in self:
            try:
                return pd.read_feather(BytesIO(self._get_s3_object_from_key(key)["Body"].read()))
            except Exception as ex:
                config.logging.warning(ex)
        return None

    def __setitem__(self, key, value):
        if isinstance(value, pd.DataFrame):
            try:
                with BytesIO() as _f:
                    value.to_feather(_f)
                    self.s3.Bucket(self.bucket_name).put_object(Key=self._s3_path(key),
                                                                Body=_f.getvalue(),
                                                                ServerSideEncryption='AES256')
            except Exception as ex:
                config.logging.warning(ex)

    def __contains__(self, key):
        try:
            self.s3.Object(self.bucket_name, self._s3_path(key)).load()
        except ClientError as ex:
            if ex.response['Error']['Code'] == "404":
                return False
            else:
                config.logging.warning(ex)
                return False
        else:
            return True
    
    def __delitem__(self, key):
        if self.delete_expired_files:
            try:
                self.s3.Object(self.bucket_name, self._s3_path(key)).delete()
            except Exception as ex:
                config.logging.warning(ex)

    def __len__(self):
        return len(self.keys())

    def keys(self):
        s3_client = boto3.client('s3')
        try:
            s3_objects = s3_client.list_objects_v2(Bucket=self.bucket_name, Prefix=self.path)['Contents']
            return list(map(lambda x: self._key(x['Key']), s3_objects))
        except Exception as ex:
            config.logging.warning(ex)
            return []
    
    def expired(self, key, ttl=None):
        if key not in self:
            return False
        if ttl is None:
            return False
        try:
            s3_object = self._get_s3_object_from_key(key)
            if isinstance(s3_object.get("LastModified", None), datetime):
                last_modified_time = s3_object["LastModified"].replace(tzinfo=None)
                if (datetime.utcnow() - last_modified_time).seconds > ttl:
                    del self[key]
                    return True
        except Exception as ex:
            config.logging.warning(ex)
            return True
        return False
        
    def _get_s3_object_from_key(self, key):
        try:
            return self.s3.Object(self.bucket_name, self._s3_path(key)).get()
        except ClientError as ex:
            return None
  
    def _s3_path(self, key):
        return f"{self.path}/{key}{self.EXTENSION}"
        
    def _key(self, s3_path):
        return s3_path.split(f"{self.path}/")[1][0:-len(self.EXTENSION)]
        



import os
import gc
import pyarrow.feather as feather
import pandas as pd
import time
from datetime import datetime

from jardin import config as config
from jardin.cache_stores.base import Base

try:
    FileExistsError
except NameError:
    FileExistsError = OSError

class Disk(Base):
    
    EXTENSION = '.feather'

    def __init__(self, dir="/tmp", limit=None):
        
        """
            dir: path of directory to use for cached files
            limit: size limit in bytes for lru cache
            ttl: cache expiry in seconds
        """
        
        self.dir = f"{dir}/jardin_cache"
        self.limit = limit
        if not os.path.exists(self.dir):
            try:
                os.makedirs(self.dir)
            except FileExistsError as ex:
                raise ex

    def __getitem__(self, key):
        if key in self:
            try:
                return feather.read_feather(self._path(key))
            except Exception as ex:
                config.logging.warning(ex)
                del self[key]
        return None

    def __setitem__(self, key, value):
        if not isinstance(value, pd.DataFrame):
            return None
        
        feather.write_feather(value, self._path(key))

        if self.limit is not None:
            if os.path.getsize(self._path(key)) > self.limit:
                raise MemoryError(f"disk cache limit exceeded by single key {key}")
            while self.size() > self.limit:
                del self[self.lru()]
            gc.collect()

    def __delitem__(self, key):
        if os.path.exists(self._path(key)):
            os.remove(self._path(key))

    def __contains__(self, key):
        return os.path.isfile(self._path(key))

    def __len__(self):
        return len(self.keys())

    def size(self):
        return sum(os.path.getsize(f) for f in self.values())

    def keys(self):
        return [self._key(f) for f in os.listdir(self.dir)]

    def values(self):
        return [os.path.join(self.dir, f) for f in os.listdir(self.dir)]

    def lru(self):
        files = sorted(self.values(), key=lambda f: os.stat(f).st_atime)
        if not files:
            return None
        return self._key(files[0])
    
    def expired(self, key, ttl=None):
        if key not in self:
            return False
        if ttl is None:
            return False
        expired = int(time.time() - os.stat(self._path(key)).st_mtime) > ttl
        if expired:
            del self[key]
            return True
        return False

    def _path(self, key):
        return os.path.join(self.dir, key + self.EXTENSION)

    def _key(self, path):
        return os.path.basename(path)[0:-len(self.EXTENSION)]



import os
from jardin import config as config
from jardin.cache_stores.disk import Disk
from jardin.cache_stores.s3 import S3

config.init()

default_cache_method = config.CACHE.get("method", None)

STORES = {}
for method, cache_config in config.CACHE.get("methods", {}).items():
    klass = None
    if method == "disk":
        klass = Disk
    elif method == "s3":
        klass = S3
    if klass is not None:
        try:
            STORES[method] = klass(**cache_config)
        except Exception as ex:
            config.logging.warning(f"Could not initialize {method} cache.")
            config.logging.warning(ex)
                
def cached(func):
    
    def wrapper(self, *args, **kwargs):
        cache = kwargs.pop('cache', False)
        cache_method = kwargs.pop('cache_method', default_cache_method)
        store = STORES.get(cache_method, None)
        
        if (store is None or not store.valid) & (cache):
            config.logging.warning("Cache store is not correctly configured!")
            return func(self, *args, **kwargs)
        
        if not cache:
            return func(self, *args, **kwargs)
        
        ttl = kwargs.pop('ttl', None)
        key = store.key(instance=self, caller=func, *args, **kwargs)
        if key not in store or store.expired(key, ttl):
            store[key] = func(self, *args, **kwargs)
        return store[key]
    
    return wrapper
